const PNG = require('pngjs').PNG;
const detect = require('detect-port');
const fs = require('fs');
const path = require('path');
const pixelmatch = require('pixelmatch');
const puppeteer = require('puppeteer');
const StaticServer = require('node-static').Server;
const util = require('util');
const http = require('http');

const readdir = util.promisify(fs.readdir);
const readFile = util.promisify(fs.readFile);
const dir = path.join(__dirname, 'scripts');
const build = path.join(__dirname, '../../public');
const generate = !!process.env.GENERATE_SCREENSHOTS;

const THRESHOLD = 10; // max mismatched pixels in screenshots

async function main() {
  const server = await serve();
  const url = `http://localhost:${server.address().port}/`;
  const browser = await puppeteer.launch();
  try {
    await testAllScripts(browser, url);
  } finally {
    server.close();
    await browser.close();
  }
}

async function serve() {
  const file = new StaticServer(build);
  const port = await detect(5000);
  const server = new http.Server((request, response) => {
    request.addListener('end', () => file.serve(request, response)).resume();
  });
  return new Promise((resolve, reject) => {
    server.listen(port, () => resolve(server));
  });
}

async function testAllScripts(browser, url) {
  const entries = await readdir(dir);
  const scripts = entries.filter(entry => entry.endsWith('.js'));
  const page = await browser.newPage();
  for (var i = 0; i < scripts.length; ++i) {
    const name = scripts[i].replace(/\.js$/, '');
    await testScript(browser, page, url, name);
  }
}

async function testScript(browser, page, url, name) {
  await page.goto(url, {waitUntil: 'networkidle'});
  const test = require(path.join(dir, name));
  try {
    await test(browser, page);
  } catch (err) {
    throw new Error(`Script "${name} failed: ${err.message}"`);
  }
  const screenshot = path.join(dir, `${name}.png`);
  const options = {
    type: 'png'
  };
  if (generate) {
    options.path = screenshot;
  }
  const gotBuffer = await page.screenshot(options);
  if (generate) {
    return;
  }
  const gotImage = PNG.sync.read(gotBuffer);
  const expectedImage = PNG.sync.read(await readFile(screenshot));
  const width = expectedImage.width;
  const height = expectedImage.height;
  const diffImage = new PNG({width, height});
  const mismatched = pixelmatch(
    gotImage.data,
    expectedImage.data,
    diffImage.data,
    width,
    height
  );
  if (mismatched > THRESHOLD) {
    throw new Error(`Screenshot ${screenshot} doesn't match`);
  }
}

if (require.main === module) {
  main().catch(err => {
    process.stderr.write(`${err.stack}\n`, () => process.exit(1));
  });
}
